
#include "expression.hpp"
#include <cmath>

using namespace std;

namespace math
{




    // untersttzte Funktionen

namespace {

double xadd (double lhs, double rhs)
{ return (lhs + rhs); }
double xsub (double lhs, double rhs)
{ return (lhs - rhs); }
double xmul (double lhs, double rhs)
{ return (lhs * rhs); }
double xdiv (double lhs, double rhs)
{
    if (rhs == 0.0)
        throw Exception (Exception::DIVISION_BY_0, "DIV: Division by 0");
    return (lhs / rhs);
}
double xlog (double lhs, double rhs)
{
    if ((lhs <= 0.0) || (rhs <= 0.0))
        throw Exception (Exception::DOMAIN, "LOG: Domain error");

    double d = log (lhs);
    if (d == 0.0)
        throw Exception (Exception::DIVISION_BY_0, "LOG: Division by 0");
    return (log (rhs) / d);
}

double xln (double rhs)
{
    if (rhs <= 0.0)
        throw Exception (Exception::DOMAIN, "LN: Domain error");
    return (log (rhs));
}

} // anonymer Namespace


    /*
     * Notation fr Ableitungsregeln:
     * fla = linker Operand
     * fra = rechter Operand (bei unren Operatoren: Argument)
     * flb = (fla)'
     * frb = (fra)'
     */

const Expression::binary_struct_t Expression::binary_functions[] = {
 /* function | name | priority | alphanumeric | derivation */
    {xadd, {"+",   0, false, "flb+frb"}},
    {xsub, {"-",   0, false, "flb-frb"}},
    {xmul, {"*",   1, false, "flb*fra+fla*frb"}},
    {xdiv, {"/",   1, false, "(flb*fra-fla*frb)/fra^2"}},
    {pow,  {"^",   2, false, "flb*fra*fla^(fra-1)+(ln fla)*fla^fra*frb"}},
    {xlog, {"log", 3, true,  "frb/fra/ln fla-(flb*ln fra/fla/(ln fla)^2"}},
};

const Expression::unary_struct_t Expression::unary_functions[] = {
 /* function | name | priority | alphanumeric | derivation */
    {xln,  {"ln",  3, true,  "frb/fra"}},
    {sin,  {"sin", 3, true,  "frb*cos fra"}},
    {cos,  {"cos", 3, true,  "frb*(-sin fra)"}},
    {tan,  {"tan", 3, true,  "frb*(1+(tan fra)^2)"}},
};

const unsigned Expression::BINARY_FUNCTION_COUNT = sizeof (Expression::binary_functions)
                                              / sizeof (Expression::binary_struct_t);
const unsigned Expression::UNARY_FUNCTION_COUNT  = sizeof (Expression::unary_functions)
                                              / sizeof (Expression::unary_struct_t);
const unsigned Expression::FUNCTION_COUNT        = BINARY_FUNCTION_COUNT + UNARY_FUNCTION_COUNT;
const unsigned Expression::UNARY_FUNCTION_OFFSET = BINARY_FUNCTION_COUNT;



void Expression::stkInit (void)
{
    fstk.resize (fstk_defsize);
    parstk.resize (parstk_defsize);
}

Expression::Expression (void)
 : varnm ("x"), vari (-1), cbfunc (NULL), cbval (NULL)
{
    fstk.clear ();
    fstk.reserve (12);
    fstk.push_back (0); // Variablenwert
    fstk.push_back (0);
    fstk.push_back (1);
    fstk.push_back (3.14159265358979323846264338327950288);
    fstk.push_back (2.71828182845904523536028747135266250);
    parstk.push_back (par_t ("pi", 3));
    parstk.push_back (par_t ("e", 4));
    drbuf.resize (FUNCTION_COUNT);
    for (unsigned n = 0; n < FUNCTION_COUNT; ++n)
        textToTerm (getFuncStruct (n).derivation, drbuf[n]);
    fstk_defsize = fstk.size ();
    parstk_defsize = parstk.size ();
}

void Expression::registerCallbackFunction (callback_t func, void* val)
{
    cbfunc = func;
    cbval = val;
}




} // namespace math
