
#include "expression.hpp"
#include <cmath>
#include <algorithm>


//#include "main_unit.h"


using namespace std;

namespace math
{


void Expression::fillBuffer (const term_t& trm)
{
    unsigned v, i = 0;

    mapbuf.clear ();
    do
    {
        v = trm[i] & NMASK;
        if ((trm[i] & TMASK) == TOPER)
            mapbuf.insert (pair <int, int> (v, i));
    } while (trm[++i] != TEND);
}

void Expression::moveBufferIndices (int i, int diff)
{
    std::multimap <int, int>::iterator p;

    p = mapbuf.begin ();
    while (p != mapbuf.end ())
    {
        if (p->second >= i)
        {
            if (p->second < i - diff)
                p->second = -1;
            else
                p->second += diff;
        }
        ++p;
    }
}

int Expression::compareOperators (unsigned op1, unsigned op2)
{
    unsigned p1, p2;

    op1 &= NMASK; op2 &= NMASK;
    p1 = op1 % FUNCTION_COUNT; p2 = op2 % FUNCTION_COUNT;
    return (op1 - p1 + getFuncStruct (p1).priority) - (op2 - p2 + getFuncStruct (p2).priority);
}
bool Expression::isConstantOperand (unsigned op)
{
    return ((op & TMASK) == TPAR) ? (parstk[op & NMASK].valoffs >= 0)
                                  : (((op & TMASK) == TVAR) ? (vari >= 0)
                                                            :  ((op & TMASK) == TNUM));
}

double Expression::getConstantValue (unsigned op)
{
    unsigned on = op & NMASK;
    return fstk[((op & TMASK) == TNUM) ? on : (((op & TMASK) == TPAR) ? parstk[on].valoffs
                                                                      : vari)];
}

void Expression::termDelete (term_t& trm, unsigned i, int n)
{
    trm.erase (trm.begin () + i, trm.begin () + i + n); // TODO: - 1 or not - 1?
    moveBufferIndices (i, -n);
}


namespace {
class ConditionalAdd : public unary_function <void, Expression::transferitem_t&>
{
private:
    unsigned i, sz;
public:
    ConditionalAdd (void)
    {}
    ConditionalAdd (unsigned i_, unsigned sz_)
     : i (i_), sz (sz_) {}
    ConditionalAdd (const ConditionalAdd& obj)
     : i (obj.i), sz (obj.sz) {}
    void operator () (Expression::transferitem_t& ti) const
    {
        if (ti.dst > i)
            ti.dst += sz;
    }
};
} // anonymer Namespace

void Expression::termExpand (term_t& base, unsigned i,
                             term_t::const_iterator src, unsigned sz)
{
    int pp = getSurrPriority (base, i) * FUNCTION_COUNT;
    unsigned v;

    base.insert (base.begin () + i, --sz, 0 /* Dummywert */);
    for (unsigned n = 0; n <= sz; ++n)
    {
        v = src[n];
        if ((v & TMASK) == TOPER)
            v += pp;
        base[i + n] = v;
    }

    for_each (trstk.begin (), trstk.end (), ConditionalAdd (i, sz));
}

void Expression::findOperands (const term_t& trm, unsigned i,
                             int& dop, int& lop, int dir)
{
    term_t::const_iterator p = trm.begin () + i;

    dop = -1;
    do
    {
        p += dir;
        if ((p < trm.begin ()) || (*p == TEND))
            break;
        if ((dop < 0) && ((*p & TMASK) == TOPER) && (compareOperators (trm[i], *p) >= 0))
            dop = p - dir - trm.begin ();
    } while (((*p & TMASK) != TOPER) || (compareOperators (trm[i], *p) <= 0));
    lop = p - dir - trm.begin ();

    if (dop < 0)
        dop = lop;
}




static const unsigned opcomm[] = {5, 1, 1, 2, 2, 4, 6, 7};
static const unsigned opcg[]   = {1, 1, 2, 2, 3, 0, 0};
static const bool     oprep[]  = {0, 1, 0, 1, 1, 0, 0};
static const unsigned oprepn[] = {1, 0, 3, 2, 2, 0, 0};



void Expression::calculateOperator (term_t& trm, unsigned li, unsigned ri, bool brkrepl)
{
    unsigned vo = (trm[ri] & NMASK) % FUNCTION_COUNT;
    double o1, o2;
    unsigned o3i;

    if (brkrepl)
        vo = oprepn[vo];

    o2 = getConstantValue (trm[ri + 1]);
    o3i = fstk.size ();

    if (vo >= UNARY_FUNCTION_OFFSET)
    {
        if (vo >= FUNCTION_COUNT)
            throw Exception (Exception::INVALID_FUNCTION,
                "Expression::calculateOperator(): Invalid function call ("
                + intToString (vo) + ")");
        trm[ri + 1] = o3i | TNUM;
        fstk.push_back ((unary_functions[vo - UNARY_FUNCTION_OFFSET].func) (o2));
        termDelete (trm, ri, 1);
    }
    else
    {
        o1 = getConstantValue (trm[li]);
        fstk.push_back ((binary_functions[vo].func) (o1, o2));
        trm[li] = o3i | TNUM;
        termDelete (trm, ri, 2);
    }
}

bool Expression::isRepOperator (const term_t& trm, unsigned i, unsigned rel_op)
{
    unsigned v = trm[i];
    return (oprep[min ((v & NMASK) % FUNCTION_COUNT, UNARY_FUNCTION_OFFSET - 1)]
         && (((v & NMASK) / FUNCTION_COUNT) == (rel_op / FUNCTION_COUNT)));
}


int Expression::getSurrPriority (const term_t& trm, unsigned i)
{
    int pr = -1;
    unsigned lpi = i, rpi = i;

    while (lpi-- > 0)
        if ((trm[lpi] & TMASK) == TOPER)
        {
            //pr = max (pr, static_cast <int> ((trm[lpi] & NMASK) / FUNCTION_COUNT));
            pr = (trm[lpi] & NMASK) / FUNCTION_COUNT;
            break;
        }

    while (trm[++rpi] != TEND)
        if ((trm[rpi] & TMASK) == TOPER)
        {
            pr = max (pr, static_cast <int> ((trm[rpi] & NMASK) / FUNCTION_COUNT));
            break;
        }
    return (pr + 1);
}

void Expression::evaluateOperator (term_t& base, int i)
{
    term_t::iterator p;
    term_t::iterator trm = base.begin () + i;

    unsigned v = *trm & NMASK;
    unsigned vo = v % FUNCTION_COUNT;
    unsigned vp = v / FUNCTION_COUNT;
    bool rep = false;

    bool no_op_after = false; // ? goto 1
    bool no_op_before = false;  // ? goto 2
    bool do_simplification = false; // ? goto 9
    bool op_exists = true;
    bool cv_is_0, cv_is_1;

    int dol, lol, dor, lor;
    bool cl, cr; // linker/rechter Wert konstant?

    float cv; // double -> float umgeht die Ungenauigkeit


    doCallback (CT_EVAL_SIMPLIFY_STEP, vp);


        // folgt hherer Operator oder konstanter Wert?
    p = trm + 1;
    if (!isConstantOperand (*p))
        do_simplification = true;
    else
    {
        while ((*p & TMASK) != TOPER)
            if (*(p++) == TEND)
            {
                no_op_after = true;
                break;
            }
        if (!no_op_after)
            if (compareOperators (*trm, *p) < 0)
                do_simplification = true;
    }


        // Nachfolger eines hheren Operators oder konstanten Wertes?
    //:1
    if (!do_simplification)
    {
        if (vo >= UNARY_FUNCTION_OFFSET)
            no_op_before = true;
        else
        {
            p = trm;
            if (!isConstantOperand (p[-1]))
                do_simplification = true;
            else
            {
                do
                    if (p-- <= base.begin ())
                    {
                        no_op_before = true;
                        break;
                    }
                while ((*p & TMASK) != TOPER);
                if (!no_op_before)
                {
                    if (compareOperators (*trm, *p) < 0)
                        do_simplification = true;
                    else
                    {
                        if ((vo > 4) || (compareOperators (*trm, *p) > 0))
                            no_op_before = true;
                        else
                            rep = isRepOperator (base, p - base.begin (), v);
                    }
                }
            }
        }
    }

    //:2
    if (!do_simplification) // beide Parameter sind konstante Werte
        calculateOperator (base, i - 1, i, rep);
    else if (vo < UNARY_FUNCTION_OFFSET)   //:9 // Vereinfachungen derzeit nur fr binre Operatoren
    {
      findOperands (base, i, dol, lol, -1);
      findOperands (base, i, dor, lor, 1);
        cl = (dol == (i - 1)) && isConstantOperand (base[dol]);
        cr = (dor == (i + 1)) && isConstantOperand (base[dor]);

        if (cl || cr)
        {
            cv = static_cast <float> (getConstantValue (base[cl ? dol : dor]));
            cv_is_0 = (cv == 0.0f);
            cv_is_1 = (cv == 1.0f);

            switch (vo)
            {
            case 0: // +
                if (cv_is_0)
                {
                    if (cl)   // 0 + c -> c
                        termDelete (base, dol, 2), op_exists = false;
                    else if (cr)   // c + 0 -> c
                        termDelete (base, i, 2), op_exists = false;
                }
                break;

            case 1: // -
                if (cv_is_0)
                    if (cr)   // c - 0 -> c
                        termDelete (base, i, 2), op_exists = false;
                break;

            case 2: // *
                if (cv_is_1)
                {
                    if (cl)   // 1 * c -> c
                        termDelete (base, dol, 2), op_exists = false;
                    else if (cr)   // c * 1 -> c
                        termDelete (base, i, 2), op_exists = false;
                }
                else if (cv_is_0)
                {   // c * 0 -> 0
                    termDelete (base, lol, lor - lol);
                    base[lol] = TNUM | registerDoubleVal (0.0);
                    op_exists = false;
                }
                break;
            case 3: // /
                if (cv_is_1)
                    if (cr)   // c / 1 -> c
                        termDelete (base, i, 2), op_exists = false;
                else if (cv_is_0)
                {
                    if (cl)   // 0 / c -> 0
                    {
                        termDelete (base, lol, lor - lol);
                        base[lol] = TNUM | registerDoubleVal (0.0);
                        op_exists = false;
                    }
                    else if (cr)   // c / 0 -> ERROR
                        throw Exception (Exception::DIVISION_BY_0, "DIV: Division by 0");
                }
                break;

            case 4: // ^
                if (cl && (cv_is_0 || cv_is_1)) // 0 ^ c -> 0, 1 ^ c -> 1
                    termDelete (base, i, dor + 1 - i), op_exists = false;
                else if (cv_is_1)  // c ^ 1 -> c
                    termDelete (base, i, 2), op_exists = false;
                else if (cv_is_0)  // c ^ 0 -> 1
                {
                    termDelete (base, dol, dor - dol);
                    base[dol] = TNUM | registerDoubleVal (1.0);
                    op_exists = false;
                }
                break;

            case 5: // log (l: base, r: arg)
                if (cr && (cv_is_1)) // b log 1 -> 0
                {
                    termDelete (base, lol, lor - lol);
                    base[lol] = TNUM | registerDoubleVal (0.0);
                    op_exists = false;
                }
                break;
            }

            if (!cr && op_exists)
            { // finde nachfolgende gleichrangige Operatoren mit konstanten Operanden
                for (int n = i + 2; n <= lor; ++n)
                {
                    if ((base[n] & TMASK) != TOPER)
                        continue;
                    if (compareOperators (base[n], *trm) > 0)
                        continue;
                    if (!isConstantOperand (base[n + 1]))
                        continue;
                    if (base[n + 2] != TEND)
                        if (compareOperators (base[n + 2], *trm) > 0)
                            continue;
                    calculateOperator (base, i - 1, n, isRepOperator (base, i - 2, *trm));
                }
            }
        }
    }
}


void Expression::removeParen (term_t::iterator p1, term_t::iterator p2,
                           unsigned pr, unsigned pi, unsigned og, bool rv)
{
    int op;
    for (term_t::iterator p = p1; p <= p2; ++p)
    {
        if ((*p & TMASK) != TOPER)
            continue;
        *p -= pi * FUNCTION_COUNT;
        op = (*p & NMASK) % FUNCTION_COUNT;
        if ((op > 4) || !rv)
            continue;
        if (((pr - pi) == ((*p & NMASK) / FUNCTION_COUNT)) && (og == opcg[op]))
            *p = TOPER | oprepn[op] + (*p & NMASK) - op;
    }
}

void Expression::evaluateParen (term_t& trm, unsigned i, unsigned pr)
{
    term_t::iterator p = trm.begin () + i;
    term_t::iterator p1 = p;
    term_t::iterator p2 = p - 1;
    unsigned op = ~0; // maximal mglicher Wert
    unsigned oo = FUNCTION_COUNT;
    unsigned cp;
    unsigned prd;
    bool rv;
    int op1, op2, cop;

    do
    {
        if ((*p & TMASK) != TOPER)
            continue;
        cp = (*p & NMASK) / FUNCTION_COUNT;
        if (cp < pr)
        {
            if (p2 == (p1 - 1))
                return;
            break;
        }
        op = min (op, cp);
        if (cp == pr)
            oo = min (oo, (*p & NMASK) % FUNCTION_COUNT);
        p2 = p;
    } while (*(++p) != TEND);

    prd = op - pr;
    rv = false;
    op1 = (p1 - 2 < trm.begin ()) ? -1 : ((p1[-2] & NMASK) % FUNCTION_COUNT);
    op2 = (p2[2] == 0) ? -1 : ((p2[2] & NMASK) % FUNCTION_COUNT);
    cop = max (op1, op2);
    if ((cop < static_cast <int> (UNARY_FUNCTION_OFFSET)) && (oo <= 4)
     && (static_cast <unsigned> (getFuncStruct (oo).priority) >= (opcomm[cop + 1] - 1)))
    {
        ++prd;
        if ((cop == op1) && (cop >= 0))
            rv = oprep[cop];
    }

    removeParen (p1, p2, op, prd,
        opcomm[min (cop + 1, static_cast <int> (UNARY_FUNCTION_OFFSET + 1))], rv);
}


void Expression::doEval (term_t& trm)
{
    term_t::iterator p = trm.begin ();
    map_t::reverse_iterator bufp;
    unsigned pr = 0;
    unsigned pa;

    if (trm[0] == TEND)
        return;

    mapbuf.clear ();
        // Klammern enumerieren
    do
    {
        if ((*p & TMASK) != TOPER)
            continue;
        pa = (*p & NMASK) / FUNCTION_COUNT;
        while (pa > pr)
            mapbuf.insert (pair <int, int> (++pr, p - trm.begin ()));
        pr = pa;
    } while (*(++p) != TEND);

    doCallback (CT_EVAL_PAREN);
    bufp = mapbuf.rbegin ();
        // irrelevante Klammern entfernen
    while (bufp != mapbuf.rend ())
    {
        evaluateParen (trm, bufp->second, bufp->first);
        ++bufp;
    }

    doCallback (CT_EVAL_SIMPLIFY);
        // vereinfachen
    fillBuffer (trm);
    bufp = mapbuf.rbegin ();
    while (bufp != mapbuf.rend ())
    {
        if ((bufp->first < 0) || (bufp->second < 0))
        {
            ++bufp;
            continue;
        }
        evaluateOperator (trm, bufp->second);
        //FrmMain->MmoOutput->Lines->Add (toText ().c_str ());
        ++bufp;
    }
}



void Expression::doDerive (term_t& trm)
{
    unsigned opsz[2];
    term_t::iterator opsrc[2];
    term_t::iterator p, p2, op1, op2, opr;
    unsigned ov1, ov2;
    transferitem_t ti;
    int psai;
    unsigned po, pt, pv, sz;
    term_t src = trm;
    unsigned cv;

    if (trm[0] == TEND)
        return;

    trm.resize (2);
    trm[0] = TSUBD | 0;
    trm[1] = TEND;
    trstk.push_back (transferitem_t (0, 0,
        find (src.begin (), src.end (), TEND) - src.begin () - 1));

    while (trstk.size ())
    {
        doCallback (CT_DERIVE_STEP, trstk.size ());

            // 1. niedrigsten Operator in Top Of Stack finden
        ti = trstk.back ();
        trstk.pop_back ();

        if (ti.from == ti.to)
        {
            trm[ti.dst] = TNUM | registerDoubleVal (((src[ti.from] & TMASK) == TVAR) ? 1.0 : 0.0);
            continue;
        }

        op1 = p = src.begin () + ti.from;
        op2 = p2 = src.begin () + ti.to;
        ov1 = ov2 = ~0;

        do
        {
            if ((*p & TMASK) != TOPER)
                continue;
            cv = *p & NMASK;
            if (cv < ov1)
            {
                ov1 = cv;
                op1 = p;
            }
            if (cv <= ov2)
            {
                ov2 = cv;
                op2 = p;
            }
        } while (++p < p2);

        opr = op2; // alle binren Funktionen sind linksassoziativ

            // 2. Funktion/Operator ableiten
        psai = ti.dst;
        po = (*opr & NMASK) % FUNCTION_COUNT;
        sz = find (drbuf[po].begin (), drbuf[po].end (), TEND) - drbuf[po].begin ();
        termExpand (trm, psai, drbuf[po].begin (), sz);


            // 3. TSUB expandieren, TSUBD auf Transfer-Stack schieben
        opsz[0] = (opr - src.begin ()) - ti.from;
        opsrc[0] = src.begin () + ti.from;
        opsz[1] = ti.to - (opr - src.begin ());
        opsrc[1] = opr + 1;

        for (int i = psai + sz - 1; i >= psai; --i)
        {
            pt = trm[i] & TMASK;
            pv = trm[i] & NMASK;
            if (pt == TSUB)
                termExpand (trm, i, opsrc[pv], opsz[pv]);
            else if (pt == TSUBD)
            {
                unsigned f = opsrc[pv] - src.begin ();
                trstk.push_back (transferitem_t (i, f, f + opsz[pv] - 1));
            }
        }
    }

    //FrmMain->MmoOutput->Lines->Add (toText ().c_str ());
    evaluate ();
}


Expression Expression::getEvaluation (void)
{
    Expression rv = *this;
    rv.evaluate ();
    return rv;
}

Expression Expression::getDerivation (void)
{
    Expression rv = *this;
    rv.doDerive (rv.term);
    return rv;
}

void Expression::evaluate (void)
{
    doCallback (CT_EVAL);
    for (unsigned n = 0; n < EVAL_PASS; ++n)
        doEval (term);
    term.resize (find (term.begin (), term.end (), TEND) - term.begin () + 1);
}

void Expression::derive (void)
{
    doCallback (CT_DERIVE);
    doDerive (term);
}


} // namespace math
