
#include "expression.hpp"
#include <cstring>
#include <cmath>
#include <cstdlib>
#include <cctype>
#include <cstdio>

//#include <windows.h>


using namespace std;

namespace math
{


class Lexer
{
    typedef bool (*isfunc_t) (char ch);

    static bool isNumeric (char ch);
    static bool isAlphaNumeric (char ch);
    static bool isWhitespace (char ch);

    static const char     opchar[];

public:
    enum ltype_t {
        LNONE,
        LNUM,
        LTOK,
        LOPER,
        LPAREN,
        LDELIM,
        LEND,
    };

private:
    string str;
    unsigned pos;
    ltype_t type, last_type;
    int val;
    double fval;
    string sval;

    string extractString (isfunc_t isf);

public:
    int getTokenAsInt (void)       { return val; }
    double getTokenAsDouble (void) { return fval; }
    void setDoubleToken (double val) { fval = val; }
    std::string getToken (void)    { return sval; }
    ltype_t getTokenType (void)    { return type; }
    ltype_t getPrevType (void)     { return last_type; }
    void parseNextToken (void);
    void setParseString (const std::string& s);
};


const char            Lexer::opchar[] = "+-*/^()";
static const char*    pfds = "flaflbfrafrb";
static const unsigned pfdv[] = {
    Expression::TSUB  | 0,
    Expression::TSUBD | 0,
    Expression::TSUB  | 1,
    Expression::TSUBD | 1,
};

string Lexer::extractString (isfunc_t isf)
{
    string s;
    char c;
    s.reserve (12);
    while (isf (c = str[pos])) // isf('\0') darf nicht true sein!
        s += tolower (c), ++pos;
    return s;
}

void Lexer::setParseString (const std::string& s)
{
    str = s;
    pos = 0;
    val = 0; fval = 0; sval.clear (); type = last_type = LNONE;
}

string doubleToString (double d, const char* format)
{
    char buf[65];
    sprintf (buf, "%g", d);
    return string (buf);
}
string intToString (int i, unsigned radix)
{
    char buf[33];
    return string (itoa (i, buf, radix));
}


static string duplicateChar (char c, unsigned cnt)
{
    string s;
    s.resize (cnt);
    for (unsigned i = 0; i < cnt; s[i++] = c)
        ;
    return s;
}


inline bool Lexer::isNumeric (char ch)
{
    static bool e = false;
    bool retval = false;

    if ((ch == 'e') || (ch == 'E'))
    {
        e = true;
        return true;
    }
    else if (e)
    {
        e = false;
        if ((ch == '+') || (ch == '-'))
            return true;
    }

    return (((ch >= '0') && (ch <= '9')) || (ch == '.'));
}
inline bool Lexer::isAlphaNumeric (char ch)
{
    return (((ch >= '0') && (ch <= '9'))
         || ((ch >= 'A') && (ch <= 'Z'))
         || ((ch >= 'a') && (ch <= 'z')));
}
inline bool Lexer::isWhitespace (char ch)
{
    return ((ch == ' ') || (ch == '\t'));
}

void Lexer::parseNextToken (void)
{
    char c;
    unsigned i;
    const char* cp;

    last_type = type;
    type = LNONE;

    while (isWhitespace (str[pos]))
        ++pos;

    c = tolower (str[pos]);

    if (c == '\0')
        type = LEND;
    else if ((c >= '0') && (c <= '9'))
    {
        fval = atof (extractString (isNumeric).c_str ());
        type = LNUM;
    }
    else if ((c >= 'a') && (c <= 'z'))
    {
        sval = extractString (isAlphaNumeric);
        type = LTOK;
        for (i = 0; i < Expression::BINARY_FUNCTION_COUNT; ++i)
            if (Expression::binary_functions[i].fs.name == sval)
            {
                type = LOPER;
                val = i;
            }
        for (i = 0; i < Expression::UNARY_FUNCTION_COUNT; ++i)
            if (Expression::unary_functions[i].fs.name == sval)
            {
                type = LOPER;
                val = i + Expression::UNARY_FUNCTION_OFFSET;
            }
    }
    else
    {
        cp = strchr (opchar, c);
        ++pos;
        if (cp)
        {
            i = cp - opchar;
            if (i <= 4)
            {
                type = LOPER;
                val = i;
            }
            else if ((i == 5) || (i == 6))
            {
                type = LPAREN;
                val = (i == 5) ? Expression::FUNCTION_COUNT
                               : -static_cast <int> (Expression::FUNCTION_COUNT);
            }
        }
    }
    if (type == LNONE)
    {
        throw Exception (Exception::SYNTAX, "Lexer::parseNextToken(): Syntax error \""
            + str.substr (pos - 1, min (str.size () - pos + 1, 20U))
            + "\" (" + intToString (pos) + ")");
    }
}


unsigned Expression::registerDoubleVal (double val)
{
    /*for (int n = fstk.size () - 1; n > 0; --n) // fstk[0] aussparen, da Variablenwert
        if (fstk[n] == val)
            return n;*/ // zu langsam
    if (val == 0.0)
        return 1;
    if (val == 1.0)
        return 2;
    fstk.push_back (val);
    return (fstk.size () - 1);
}

unsigned Expression::registerParam (const std::string& name)
{
    for (int n = parstk.size() - 1; n >= 0; --n)
        if (parstk[n].name == name)
            return n;
    parstk.push_back (par_t (name, -1));
    return (parstk.size () - 1);
}

void Expression::registerParamValue (const std::string& name, double val)
{
    parstk[registerParam (name)].valoffs = registerDoubleVal (val);
}

void Expression::removeParamValue (const std::string& name)
{
    parstk[registerParam (name)].valoffs = -1;
}




    // lexikalische Analyse
void Expression::textToTerm (const std::string& text, term_t& trm)
{
    vector <unsigned>::iterator dst = trm.begin ();
    Lexer lex;
    Lexer::ltype_t lle = Lexer::LOPER;
    bool neg = false;
    int pr = 0;
    bool cont = true;

    doCallback (CT_FROMTEXT, 0);
    lex.setParseString (text);
    do
    {
        if ((trm.end () - dst) < 5)
        {
            int offs = dst - trm.begin ();
            trm.resize (trm.size () + 20);
            dst = trm.begin () + offs;
        }
        lex.parseNextToken ();
        switch (lex.getTokenType ())
        {
        case Lexer::LNUM:
            if (Expression::BIND_NEGATIVE && neg)
            {
                dst -= 2;
                lex.setDoubleToken (-lex.getTokenAsDouble ());
            }
            neg = false;
            lle = Lexer::LNUM;
            *(dst++) = TNUM | registerDoubleVal (lex.getTokenAsDouble ());
            break;

        case Lexer::LOPER:
            if ((lex.getTokenAsInt () == 1) && (lle == Lexer::LOPER))
            {
                *(dst++) = TNUM; // "-" -> "0-"
                neg = true;
            }
            *(dst++) = TOPER | (pr + lex.getTokenAsInt ());
            lle = Lexer::LOPER;
            break;

        case Lexer::LTOK:
            if (lex.getToken () == varnm)
                *dst = TVAR;
            else
            {
                const char* p = strstr (pfds, lex.getToken ().c_str ());
                unsigned i = (p - pfds);
                if (p && (lex.getToken ().size () == 3) && (i % 3 == 0))
                    *dst = pfdv[i / 3];
                else
                    *dst = TPAR | registerParam (lex.getToken ());
            }
            ++dst;
            lle = Lexer::LTOK;
            neg = false;
            break;

        case Lexer::LPAREN:
            pr += lex.getTokenAsInt ();
            neg = false;
            break;

        case Lexer::LEND:
            cont = false;
            break;

        default:
            throw Exception (Exception::SYNTAX, "Expression::textToTerm(): Syntax error \""
                + lex.getToken () + "\"");
        }

    } while (cont);

    *dst = 0;

    /*std::string dbgstr;
    char nbuf[11];
    for (int i = 0; trm[i]; ++i)
    {
        sprintf (nbuf, "%X ", trm[i]);
        dbgstr += nbuf;
    }
    MessageBox (NULL, dbgstr.c_str (), "", 0);*/

}

void Expression::readFromText (const std::string& text)
{
    textToTerm (text, term);
}

std::string Expression::toText (const char* format)
{
    unsigned pr = 0;
    string dst;
    bool tok = false;
    bool lastnum0 = false;
    unsigned l = 0;
    unsigned v, i;
    vector <unsigned>::iterator src = term.begin ();

    doCallback (CT_TOTEXT, 0);
    while (*src != TEND)
    {
        v = *src & NMASK;
        switch (*src & TMASK)
        {
        case TNUM:
            if (tok)
                dst += ' ';
            tok = false;
            l = dst.size ();
            dst += doubleToString (fstk[v], format);
            lastnum0 = (fstk[v] == 0.0);
            break;

        case TOPER:
            i = v / FUNCTION_COUNT;
            v = v % FUNCTION_COUNT;
            if (lastnum0 && (v == 1))
                dst = dst.substr (0, dst.size () - 1); // "0-" -> "-"
            if ((tok || (v < UNARY_FUNCTION_OFFSET)) && getFuncStruct (v).alpha && (i == pr))
                dst += ' ';
            if (v >= UNARY_FUNCTION_OFFSET)
                l = dst.size ();
            if (i > pr)
                dst = dst.substr (0, l)
                    + duplicateChar ('(', i - pr)
                    + dst.substr (l);
            if (i < pr)
                dst += duplicateChar (')', pr - i);
            pr = i;
            dst += getFuncStruct (v).name;
            tok = getFuncStruct (v).alpha;
            break;

        case TPAR:
            lastnum0 = false;
            if (tok)
                dst += ' ';
            tok = true;
            l = dst.size ();
            dst += parstk[v].name;
            break;

        case TVAR:
            lastnum0 = false;
            if (tok)
                dst += ' ';
            tok = true;
            l = dst.size ();
            dst += varnm;
            break;


            // nur fr Debuggingzwecke - sollte eigentlich nicht auftauchen
        case TSUB:
            lastnum0 = false;
            if (tok)
                dst += ' ';
            tok = true;
            l = dst.size ();
            dst += (v == 0) ? "fla" : "fra";
            break;

        case TSUBD:
            lastnum0 = false;
            if (tok)
                dst += ' ';
            tok = true;
            l = dst.size ();
            dst += (v == 0) ? "flb" : "frb";
            break;
        }
        ++src;
    }

    return (dst + duplicateChar (')', pr));
}

double Expression::toDouble (void)
{
    if (!isConstantValue ())
        throw Exception (Exception::NOT_A_CONSTANT,
        "Expression::toDouble(): Not a constant value!");
    return (fstk[term[0] & NMASK]);
}
bool Expression::isConstantValue (void)
{
    return (((term[0] & TMASK) == TNUM) && (term[1] == TEND));
}





} // namespace math
