//---------------------------------------------------------------------------

#include <vcl.h>
#include <process.h>
#pragma hdrstop

#include "main_unit.h"
#include "expression.hpp"
#include "taylordlg_unit.h"
#include "statusdlg_unit.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFrmMain *FrmMain;



inline unsigned long long getFactorial (unsigned long long val)
{
    return ((val <= 1) ? 1 : (val * getFactorial (val - 1)));
}


void myCallBackFunction (math::Expression::callback_task_t task, unsigned val, void* data)
{
    cbfuncdata_t* cbfd = (cbfuncdata_t*) data;

    switch (task)
    {
    case math::Expression::CT_DERIVE_STEP:        // param

        break;
    case math::Expression::CT_EVAL_SIMPLIFY_STEP: // param
        if (val == cbfd->priority)
            return;
        if (cbfd->priority == ~0)
        {
            cbfd->max_pr = val;
            cbfd->PbrStatus->Max = val;
        }
        cbfd->priority = val;
        cbfd->PbrStatus->Position = cbfd->max_pr - val;
        break;

    case math::Expression::CT_EVAL:
        cbfd->MmoOutput->Lines->Add ("Evaluieren...");
        cbfd->priority = 0;
        break;

    case math::Expression::CT_EVAL_PAREN:
        cbfd->MmoOutput->Lines->Add ("  Klammern auflsen...");
        break;

    case math::Expression::CT_EVAL_SIMPLIFY:
        cbfd->MmoOutput->Lines->Add ("  Vereinfachen...");
        cbfd->priority = ~0;
        break;

    case math::Expression::CT_DERIVE:
        cbfd->MmoOutput->Lines->Add ("Ableiten...");
        break;

    case math::Expression::CT_TOTEXT:
        cbfd->MmoOutput->Lines->Add ("In Text konvertieren...");
        break;

    case math::Expression::CT_FROMTEXT:
        cbfd->MmoOutput->Lines->Add ("In Formel konvertieren...");
        break;

    }
}


void threadFunc (void* data);

void doTask (task_t task, double xoffs = 0.0, unsigned degree = 0)
{
    switch (task)
    {
    case TASK_EVAL:
        FrmStatusDlg->LblAction->Caption = "Evaluieren";
        break;

    case TASK_DERIVE:
        FrmStatusDlg->LblAction->Caption = "Ableiten";
        break;

    case TASK_TAYLOR:
        FrmStatusDlg->LblAction->Caption = "Taylor-Approximation";
        break;
    }

    FrmStatusDlg->MmoStatus->Clear ();

    cbfuncdata_t cbfd;
    cbfd.MmoOutput = FrmStatusDlg->MmoStatus;
    cbfd.PbrStatus = FrmStatusDlg->PbrProgress;
    cbfd.priority = ~0;
    cbfd.task = task;
    cbfd.done = false;
    cbfd.xoffs = xoffs;
    cbfd.degree = degree;

    cbfd.hThread = (HANDLE) (_beginthread (threadFunc, 0, &cbfd));
    FrmStatusDlg->cbfd = &cbfd;
    FrmStatusDlg->ShowModal ();
    
    FrmMain->MmoOutput->Lines->Add (FrmMain->lastExpression);
    FrmMain->MmoInput->SelectAll ();
    FrmMain->MmoInput->SetFocus ();
}

void threadFunc (void* data)
{
    cbfuncdata_t* cbfd = (cbfuncdata_t*) data;
    math::Expression frm, res;

    unsigned degree, i;
    double xoffs;
    std::string str;

    frm.registerCallbackFunction (myCallBackFunction, cbfd);

    
    try
    {
        switch (cbfd->task)
        {
        case TASK_EVAL:
            frm.readFromText (FrmMain->MmoInput->Text.c_str ());
            if (!FrmMain->EdtXValue->Text.IsEmpty ())
                    frm.registerVariableValue (FrmMain->EdtXValue->Text.ToDouble ());
            FrmMain->lastExpression = frm.getEvaluation ().toText ().c_str ();
            break;

        case TASK_DERIVE:
            frm.readFromText (FrmMain->MmoInput->Text.c_str ());
            FrmMain->lastExpression = frm.getDerivation ().toText ().c_str ();
            break;

        case TASK_TAYLOR:
            degree = cbfd->degree;
            xoffs = cbfd->xoffs;

            frm.readFromText (FrmMain->MmoInput->Text.c_str ());

            for (i = 0; i <= degree; ++i)
            {
                math::Expression cal = frm;
                cal.registerVariableValue (xoffs);
                cal.evaluate ();
                str += "(" + cal.toText () + ")/"
                    + math::doubleToString (static_cast <double> (getFactorial (i)))
                    + "*(x-" + math::doubleToString (xoffs)
                    + ")^" + math::intToString (i) + "+";
                frm.derive ();
            }
            str.resize (str.size () - 1); // letztes "+" entfernen
            res.readFromText (str);
            res.evaluate ();
            FrmMain->lastExpression = res.toText ().c_str ();
            break;
        }
    }
    catch (math::Exception& ex)
    {
        Application->MessageBox (ex.getMessage ().c_str (),
            "Exception occured in math::Expression", MB_ICONWARNING);
    }

    cbfd->done = true;

    FrmStatusDlg->Close ();
    
    _endthread ();
}

//---------------------------------------------------------------------------
__fastcall TFrmMain::TFrmMain(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnEvalClick(TObject *Sender)
{
    doTask (TASK_EVAL);
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnDeriveClick(TObject *Sender)
{
    doTask (TASK_DERIVE);
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnCloseClick(TObject *Sender)
{
    Close ();
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnAboutClick(TObject *Sender)
{
    Application->MessageBox (
        "math::Expression v0.2.0\n"
        "(c) 2007 by Moritz Beutel\n\n"
        "Dieses kleine Programm demonstriert die Verwendung der math::Expression-Klasse.\n\n"
        "Erstellt mit Borland Turbo C++ 2006.",
        "Info ber math::Expression", MB_ICONINFORMATION);        
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnClearInputClick(TObject *Sender)
{
    MmoInput->Clear ();
    MmoInput->SetFocus ();        
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnClearOutputClick(TObject *Sender)
{
    MmoOutput->Clear ();        
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnUseResultClick(TObject *Sender)
{
    if (!lastExpression.IsEmpty ())
    {
        MmoInput->Text = lastExpression;
        MmoInput->SelectAll ();
        MmoInput->SetFocus ();
    }
}

//---------------------------------------------------------------------------

void __fastcall TFrmMain::BtnTaylorClick(TObject *Sender)
{
    double xoffs;
    unsigned degree;

    degree = FrmTaylorDialog->getTaylorPolynomDegree (xoffs);
    if (!degree)
        return;

    doTask (TASK_TAYLOR, xoffs, degree);
}
//---------------------------------------------------------------------------

